@@ -23,14 +23,10 @@
 #include <stdlib.h>
 #include <string.h>
 
-#ifndef DISABLE_REGEX
-#include <sys/types.h>
-#include <regex.h>
-#endif /* !DISABLE_REGEX */
-
 #include "gpg.h"
 #include "../common/status.h"
 #include "../common/iobuf.h"
+#include "../regexp/jimregexp.h"
 #include "keydb.h"
 #include "../common/util.h"
 #include "options.h"
@@ -41,6 +37,10 @@
 #include "tdbio.h"
 #include "trustdb.h"
 #include "tofu.h"
+#include "key-clean.h"
+
+static void write_record (ctrl_t ctrl, TRUSTREC *rec);
+static void do_sync(void);
 
 
 typedef struct key_item **KeyHashTable; /* see new_key_hash_table() */
@@ -192,7 +192,7 @@ release_key_array ( struct key_array *keys )
  * before initializing the validation module.
  * FIXME: Should be replaced by a function to add those keys to the trustdb.
  */
-void
+static void
 tdb_register_trusted_keyid (u32 *keyid)
 {
   struct key_item *k;
@@ -204,22 +204,35 @@ tdb_register_trusted_keyid (u32 *keyid)
   user_utk_list = k;
 }
 
+
 void
-tdb_register_trusted_key( const char *string )
+tdb_register_trusted_key (const char *string)
 {
   gpg_error_t err;
   KEYDB_SEARCH_DESC desc;
+  u32 kid[2];
 
   err = classify_user_id (string, &desc, 1);
-  if (err || desc.mode != KEYDB_SEARCH_MODE_LONG_KID )
+  if (!err)
     {
-      log_error(_("'%s' is not a valid long keyID\n"), string );
-      return;
+      if (desc.mode == KEYDB_SEARCH_MODE_LONG_KID)
+        {
+          tdb_register_trusted_keyid (desc.u.kid);
+          return;
+        }
+      if (desc.mode == KEYDB_SEARCH_MODE_FPR
+          || desc.mode == KEYDB_SEARCH_MODE_FPR20)
+        {
+          kid[0] = buf32_to_u32 (desc.u.fpr+12);
+          kid[1] = buf32_to_u32 (desc.u.fpr+16);
+          tdb_register_trusted_keyid (kid);
+          return;
+        }
     }
-
-  register_trusted_keyid(desc.u.kid);
+  log_error (_("'%s' is not a valid long keyID\n"), string );
 }
 
+
 /*
  * Helper to add a key to the global list of ultimately trusted keys.
  * Returns: true = inserted, false = already in list.
@@ -244,6 +257,49 @@ add_utk (u32 *kid)
 }
 
 
+/* Add/remove KID to/from the list of ultimately trusted keys.  */
+void
+tdb_update_utk (u32 *kid, int add)
+{
+  struct key_item *k, *k_prev;
+
+  k_prev = NULL;
+  for (k = utk_list; k; k = k->next)
+    if (k->kid[0] == kid[0] && k->kid[1] == kid[1])
+      break;
+    else
+      k_prev = k;
+
+  if (add)
+    {
+      if (!k)
+        {
+          k = new_key_item ();
+          k->kid[0] = kid[0];
+          k->kid[1] = kid[1];
+          k->ownertrust = TRUST_ULTIMATE;
+          k->next = utk_list;
+          utk_list = k;
+          if ( opt.verbose > 1 )
+            log_info(_("key %s: accepted as trusted key\n"), keystr(kid));
+        }
+    }
+  else
+    {
+      if (k)
+        {
+          if (k_prev)
+            k_prev->next = k->next;
+          else
+            utk_list = NULL;
+
+          xfree (k->trust_regexp);
+          xfree (k);
+        }
+    }
+}
+
+
 /****************
  * Verify that all our secret keys are usable and put them into the utk_list.
  */
@@ -253,10 +309,11 @@ verify_own_keys (ctrl_t ctrl)
   TRUSTREC rec;
   ulong recnum;
   int rc;
-  struct key_item *k;
+  struct key_item *k, *k2;
+  int need_revalidation = 0;
 
   if (utk_list)
-    return;
+    return; /* Has already been run.  */
 
   /* scan the trustdb to find all ultimately trusted keys */
   for (recnum=1; !tdbio_read_record (recnum, &rec, 0); recnum++ )
@@ -264,24 +321,47 @@ verify_own_keys (ctrl_t ctrl)
       if ( rec.rectype == RECTYPE_TRUST
            && (rec.r.trust.ownertrust & TRUST_MASK) == TRUST_ULTIMATE)
         {
-            byte *fpr = rec.r.trust.fingerprint;
-            int fprlen;
-            u32 kid[2];
+          byte *fpr = rec.r.trust.fingerprint;
+          int fprlen;
+          u32 kid[2];
 
-            /* Problem: We do only use fingerprints in the trustdb but
-             * we need the keyID here to indetify the key; we can only
-             * use that ugly hack to distinguish between 16 and 20
-             * butes fpr - it does not work always so we better change
-             * the whole validation code to only work with
-             * fingerprints */
-            fprlen = (!fpr[16] && !fpr[17] && !fpr[18] && !fpr[19])? 16:20;
-            keyid_from_fingerprint (ctrl, fpr, fprlen, kid);
-            if (!add_utk (kid))
-	      log_info(_("key %s occurs more than once in the trustdb\n"),
-		       keystr(kid));
+          /* Problem: We do only use fingerprints in the trustdb but
+           * we need the keyID here to indetify the key; we can only
+           * use that ugly hack to distinguish between 16 and 20 bytes
+           * fpr - it does not work always so we better change the
+           * whole validation code to only work with fingerprints */
+          fprlen = (!fpr[16] && !fpr[17] && !fpr[18] && !fpr[19])? 16:20;
+          keyid_from_fingerprint (ctrl, fpr, fprlen, kid);
+          if (!add_utk (kid))
+            log_info(_("key %s occurs more than once in the trustdb\n"),
+                     keystr(kid));
+          else if ((rec.r.trust.flags & 1))
+            {
+              /* Record marked as inserted via --trusted-key.  Is this
+               * still the case?  */
+              for (k2 = user_utk_list; k2; k2 = k2->next)
+                if (k2->kid[0] == kid[0] && k2->kid[1] == kid[1])
+                  break;
+              if (!k2) /* No - clear the flag.  */
+                {
+                  if (DBG_TRUST)
+                    log_debug ("clearing former --trusted-key %s\n",
+                               keystr (kid));
+                  rec.r.trust.ownertrust = TRUST_UNKNOWN;
+                  rec.r.trust.flags &= ~(rec.r.trust.flags & 1);
+                  write_record (ctrl, &rec);
+                  need_revalidation = 1;
+                }
+            }
         }
     }
 
+  if (need_revalidation)
+    {
+      tdb_revalidation_mark (ctrl);
+      do_sync ();
+    }
+
   /* Put any --trusted-key keys into the trustdb */
   for (k = user_utk_list; k; k = k->next)
     {
@@ -290,7 +370,7 @@ verify_own_keys (ctrl_t ctrl)
           PKT_public_key pk;
 
           memset (&pk, 0, sizeof pk);
-          rc = get_pubkey (ctrl, &pk, k->kid);
+          rc = get_pubkey_with_ldap_fallback (ctrl, &pk, k->kid);
           if (rc)
 	    log_info(_("key %s: no public key for trusted key - skipped\n"),
 		     keystr(k->kid));
@@ -298,11 +378,13 @@ verify_own_keys (ctrl_t ctrl)
 	    {
 	      tdb_update_ownertrust
                 (ctrl, &pk, ((tdb_get_ownertrust (ctrl, &pk, 0) & ~TRUST_MASK)
-                             | TRUST_ULTIMATE ));
+                             | TRUST_ULTIMATE ),  1);
 	      release_public_key_parts (&pk);
 	    }
 
-          log_info (_("key %s marked as ultimately trusted\n"),keystr(k->kid));
+          if (!opt.quiet)
+            log_info (_("key %s marked as ultimately trusted\n"),
+                      keystr(k->kid));
         }
     }
 
@@ -605,7 +687,7 @@ tdb_check_or_update (ctrl_t ctrl)
       if (opt.interactive)
 	update_trustdb (ctrl);
       else if (!opt.no_auto_check_trustdb)
-	check_trustdb (ctrl);
+        check_trustdb (ctrl);
     }
 }
 
@@ -737,7 +819,8 @@ tdb_get_min_ownertrust (ctrl_t ctrl, PKT_public_key *pk, int no_create)
  * The key should be a primary one.
  */
 void
-tdb_update_ownertrust (ctrl_t ctrl, PKT_public_key *pk, unsigned int new_trust )
+tdb_update_ownertrust (ctrl_t ctrl, PKT_public_key *pk, unsigned int new_trust,
+                       int as_trusted_key)
 {
   TRUSTREC rec;
   gpg_error_t err;
@@ -749,11 +832,24 @@ tdb_update_ownertrust (ctrl_t ctrl, PKT_public_key *pk, unsigned int new_trust )
   if (!err)
     {
       if (DBG_TRUST)
-        log_debug ("update ownertrust from %u to %u\n",
-                   (unsigned int)rec.r.trust.ownertrust, new_trust );
+        log_debug ("update ownertrust from %u to %u%s\n",
+                   (unsigned int)rec.r.trust.ownertrust, new_trust,
+                   as_trusted_key? " via --trusted-key":"");
       if (rec.r.trust.ownertrust != new_trust)
         {
           rec.r.trust.ownertrust = new_trust;
+          /* Clear or set the trusted key flag if the new value is
+           * ultimate.  This is required so that we know which keys
+           * have been added by --trusted-keys.  */
+          if ((rec.r.trust.ownertrust & TRUST_MASK) == TRUST_ULTIMATE)
+            {
+              if (as_trusted_key)
+                rec.r.trust.flags |= 1;
+              else
+                rec.r.trust.flags &= ~(rec.r.trust.flags & 1);
+            }
+          else
+            rec.r.trust.flags &= ~(rec.r.trust.flags & 1);
           write_record (ctrl, &rec);
           tdb_revalidation_mark (ctrl);
           do_sync ();
@@ -764,13 +860,17 @@ tdb_update_ownertrust (ctrl_t ctrl, PKT_public_key *pk, unsigned int new_trust )
       size_t dummy;
 
       if (DBG_TRUST)
-        log_debug ("insert ownertrust %u\n", new_trust );
+        log_debug ("insert ownertrust %u%s\n", new_trust,
+                   as_trusted_key? " via --trusted-key":"");
 
       memset (&rec, 0, sizeof rec);
       rec.recnum = tdbio_new_recnum (ctrl);
       rec.rectype = RECTYPE_TRUST;
       fingerprint_from_pk (pk, rec.r.trust.fingerprint, &dummy);
       rec.r.trust.ownertrust = new_trust;
+      if ((rec.r.trust.ownertrust & TRUST_MASK) == TRUST_ULTIMATE
+          && as_trusted_key)
+        rec.r.trust.flags = 1;
       write_record (ctrl, &rec);
       tdb_revalidation_mark (ctrl);
       do_sync ();
@@ -844,6 +944,7 @@ update_min_ownertrust (ctrl_t ctrl, u32 *kid, unsigned int new_trust)
 
 /*
  * Clear the ownertrust and min_ownertrust values.
+ * Also schedule a revalidation if a stale validity record exists.
  *
  * Return: True if a change actually happened.
  */
@@ -877,6 +978,26 @@ tdb_clear_ownertrusts (ctrl_t ctrl, PKT_public_key *pk)
           do_sync ();
           return 1;
         }
+      else
+        {
+          /* Check whether we have a stale RECTYPE_VALID for that key
+           * and if its validity ist set, schedule a revalidation.  */
+          ulong recno = rec.r.trust.validlist;
+          while (recno)
+            {
+              read_record (recno, &rec, RECTYPE_VALID);
+              if (rec.r.valid.validity)
+                break;
+              recno = rec.r.valid.next;
+            }
+          if (recno)
+            {
+              if (DBG_TRUST)
+                log_debug ("stale validity value detected"
+                           " - scheduling check\n");
+              tdb_revalidation_mark (ctrl);
+            }
+        }
     }
   else if (gpg_err_code (err) != GPG_ERR_NOT_FOUND)
     {
@@ -1391,7 +1512,7 @@ ask_ownertrust (ctrl_t ctrl, u32 *kid, int minimum)
     {
       log_info("force trust for key %s to %s\n",
 	       keystr(kid),trust_value_to_string(opt.force_ownertrust));
-      tdb_update_ownertrust (ctrl, pk, opt.force_ownertrust);
+      tdb_update_ownertrust (ctrl, pk, opt.force_ownertrust, 0);
       ot=opt.force_ownertrust;
     }
   else
@@ -1504,8 +1625,7 @@ store_validation_status (ctrl_t ctrl, int depth,
 
 /* Returns a sanitized copy of the regexp (which might be "", but not
    NULL). */
-#ifndef DISABLE_REGEX
-/* Operator charactors except '.' and backslash.
+/* Operator characters except '.' and backslash.
    See regex(7) on BSD.  */
 #define REGEXP_OPERATOR_CHARS "^[$()|*+?{"
 
@@ -1569,52 +1689,51 @@ sanitize_regexp(const char *old)
 
   return new;
 }
-#endif /*!DISABLE_REGEX*/
+
 
 /* Used by validate_one_keyblock to confirm a regexp within a trust
-   signature.  Returns 1 for match, and 0 for no match or regex
-   error. */
+ * signature.  Returns 1 for match, and 0 for no match or regex
+ * error. */
 static int
-check_regexp(const char *expr,const char *string)
+check_regexp (const char *expr,const char *string)
 {
-#ifdef DISABLE_REGEX
-  (void)expr;
-  (void)string;
-  /* When DISABLE_REGEX is defined, assume all regexps do not
-     match. */
-  return 0;
-#else
   int ret;
   char *regexp;
+  char *stringbuf = NULL;
+  regex_t pat;
 
-  regexp=sanitize_regexp(expr);
+  regexp = sanitize_regexp (expr);
 
-#ifdef __riscos__
-  ret=riscos_check_regexp(expr, string, DBG_TRUST);
-#else
-  {
-    regex_t pat;
+  ret = regcomp (&pat, regexp, (REG_ICASE|REG_EXTENDED));
+  if (!ret)
+    {
+      if (*regexp == '<' && !strchr (string, '<')
+          && is_valid_mailbox (string))
+        {
+          /* The R.E. starts with an angle bracket but STRING seems to
+           * be a plain mailbox (e.g. "foo@example.org").  The
+           * commonly used R.E. pattern "<[^>]+[@.]example\.org>$"
+           * won't be able to detect this.  Thus we enclose STRING
+           * into angle brackets for checking.  */
+          stringbuf = xstrconcat ("<", string, ">", NULL);
+          string = stringbuf;
+        }
+      ret = regexec (&pat, string, 0, NULL, 0);
+      regfree (&pat);
+    }
 
-    ret=regcomp(&pat,regexp,REG_ICASE|REG_NOSUB|REG_EXTENDED);
-    if(ret==0)
-      {
-	ret=regexec(&pat,string,0,NULL,0);
-	regfree(&pat);
-      }
-    ret=(ret==0);
-  }
-#endif
+  ret = !ret;
 
-  if(DBG_TRUST)
-    log_debug("regexp '%s' ('%s') on '%s': %s\n",
-	      regexp,expr,string,ret?"YES":"NO");
-
-  xfree(regexp);
+  if (DBG_TRUST)
+    log_debug ("regexp '%s' ('%s') on '%s'%s: %s\n",
+               regexp, expr, string, stringbuf? " (fixed)":"", ret? "YES":"NO");
 
+  xfree (regexp);
+  xfree (stringbuf);
   return ret;
-#endif
 }
 
+
 /*
  * Return true if the key is signed by one of the keys in the given
  * key ID list.  User IDs with a valid signature are marked by node
